// VPTestBench.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

const int iDstBitDepth = 10;
const int iSrcBitDepth = 8;
void UYVYToYUV422P10DI_xme(PBYTE pbDstY, PBYTE pbDstU, PBYTE pbDstV, INT iDstYStride, INT iDstUvStride,
	PBYTE pbSrcY, PBYTE pbSrcU, PBYTE pbSrcV, INT iSrcYStride, INT iSrcUvStride,
	UINT uWidth, INT iHeight);

void UYVYToYUV422P10DI_avx2(PBYTE pbDstY, PBYTE pbDstU, PBYTE pbDstV, INT iDstYStride, INT iDstUvStride,
	PBYTE pbSrcY, PBYTE pbSrcU, PBYTE pbSrcV, INT iSrcYStride, INT iSrcUvStride,
	UINT uWidth, INT iHeight);

void UYVYToYUV422P10DI_avx512(PBYTE pbDstY, PBYTE pbDstU, PBYTE pbDstV, INT iDstYStride, INT iDstUvStride,
	PBYTE pbSrcY, PBYTE pbSrcU, PBYTE pbSrcV, INT iSrcYStride, INT iSrcUvStride,
	UINT uWidth, INT iHeight);

#if linux
static inline UINT __rdtsc(void) {
	UINT a = 0;

	asm volatile("rdtsc" : "=a"(a) ::"edx");
	return a;
}
#endif
INT64 GetPerfCount()
{
#ifdef WIN32
	LARGE_INTEGER count;
	QueryPerformanceCounter(&count);
	return count.QuadPart;
#else
	struct timespec time;
	clock_gettime(CLOCK_MONOTONIC, &time);
	return time.tv_sec * 1000000 + time.tv_nsec/1000;
	//return (INT64)__rdtsc();
#endif
}

void* AlignedMalloc(INT Size, INT alignment)
{
	INT64 v;
	BYTE *p, *p1;
	INT off;
	static UINT alignment_static[] = {1, 2, 4, 8, 16, 32, 64, 128};
	INT ALIGNBYTES = alignment;
	INT i = 0;
	for (i=0; i<sizeof(alignment_static)/sizeof(alignment_static[0]); i++)
	{
		if (alignment == alignment_static[i])
			break;
	}
	if (i == sizeof(alignment_static)/sizeof(alignment_static[0]))
		return NULL;

	p = (BYTE*)malloc(Size + ALIGNBYTES);
	if(p == NULL)
		return NULL;
	v = (INT64)p;
	p1 = (BYTE*)((v + ALIGNBYTES)& ~(ALIGNBYTES-1) );
	memset(p1,0,Size);
	off = p1 - p;
	p = p1 -1;
	*p = (BYTE)off;
	return p1;
}

void AlignedFree(void* ptr)
{
	BYTE *p1;
	BYTE v ;
	BYTE *p2;
	if(ptr == NULL)
		return;
	p1 = (BYTE*)ptr - 1;
	v = *p1;
	p2 = (BYTE*)ptr - v;
	free(p2);
}

void UYVYToYUV422P10DI_C(PBYTE pbDstY, PBYTE pbDstU, PBYTE pbDstV, INT iDstYStride, INT iDstUvStride,
	PBYTE pbSrcY, PBYTE pbSrcU, PBYTE pbSrcV, INT iSrcYStride, INT iSrcUvStride,
	UINT uWidth, INT iHeight)
{
	INT iSrcShift = (iDstBitDepth - iSrcBitDepth) < 0 ? 0 : (iDstBitDepth - iSrcBitDepth);
	INT iDstShift = (iSrcBitDepth - iDstBitDepth) < 0 ? 0 : (iSrcBitDepth - iDstBitDepth);

	SrcT *pSrcY = (SrcT *)pbSrcY;
	DstT *pDstY = (DstT *)pbDstY;
	DstT *pDstU = (DstT *)pbDstU;
	DstT *pDstV = (DstT *)pbDstV;

	iSrcYStride >>= sizeof(SrcT) - 1;
	iDstYStride >>= sizeof(DstT) - 1;
	iDstUvStride >>= sizeof(DstT)- 1;

	if(iHeight < 0)
	{
		iHeight = -iHeight;
		pDstY += (iHeight-1) * iDstYStride;
		iDstYStride = -iDstYStride;
		iDstUvStride = -iDstUvStride;
	}

	INT iSrcYDif  = iSrcYStride - (uWidth<<1);
	INT iDstUvDif = iDstUvStride - (uWidth>>1);
	INT iDstYDif  = iDstYStride - uWidth;
	INT iH = iHeight - 2;

	//top line
	{
		for(UINT x = 0; x < (uWidth>>1); x++)
		{
			pDstY[x*2    ] = (DstT)_BitConv(pSrcY[x*4 + 1], iSrcShift, iDstShift);
			pDstY[x*2 + 1] = (DstT)_BitConv(pSrcY[x*4 + 3], iSrcShift, iDstShift);
			pDstU[x] = (DstT)_BitConv(pSrcY[x*4 + 0], iSrcShift, iDstShift);
			pDstV[x] = (DstT)_BitConv(pSrcY[x*4 + 2], iSrcShift, iDstShift);
		}

		pDstY += iDstYStride;
		pDstU += iDstUvStride;
		pDstV += iDstUvStride;
	}

	for(INT y = 0; y < iH; y++)
	{
		for(UINT x = 0; x < (uWidth>>1); x++)
		{
			pDstU[0] = (DstT)_DI_BOB_3(pSrcY[0], pSrcY[iSrcYStride],   pSrcY[iSrcYStride*2],   iSrcShift, iDstShift);
			pDstY[0] = (DstT)_DI_BOB_3(pSrcY[1], pSrcY[iSrcYStride+1], pSrcY[iSrcYStride*2+1], iSrcShift, iDstShift);
			pDstV[0] = (DstT)_DI_BOB_3(pSrcY[2], pSrcY[iSrcYStride+2], pSrcY[iSrcYStride*2+2], iSrcShift, iDstShift);
			pDstY[1] = (DstT)_DI_BOB_3(pSrcY[3], pSrcY[iSrcYStride+3], pSrcY[iSrcYStride*2+3], iSrcShift, iDstShift);

			pDstY +=2;
			pDstU ++;
			pDstV ++;
			pSrcY +=4;
		}

		pSrcY += iSrcYDif;
		pDstU += iDstUvDif;
		pDstV += iDstUvDif;
		pDstY += iDstYDif;
	}

	//bottom line
	{
		pSrcY += iSrcYStride;

		for(UINT x = 0; x < (uWidth>>1); x++)
		{
			pDstY[x*2    ] = (DstT)_BitConv(pSrcY[x*4 + 1], iSrcShift, iDstShift);
			pDstY[x*2 + 1] = (DstT)_BitConv(pSrcY[x*4 + 3], iSrcShift, iDstShift);
			pDstU[x] = (DstT)_BitConv(pSrcY[x*4 + 0], iSrcShift, iDstShift);
			pDstV[x] = (DstT)_BitConv(pSrcY[x*4 + 2], iSrcShift, iDstShift);
		}
	}
}

int main(int argc, char *argv[])
{
	//compare result

	int iWidth = 3840;
	int iHeight = 2160;
	int iInSize = iWidth * iHeight * 2;
	int iOutSize =  iWidth * iHeight * 2 * 2;
	int iFrames = 2000;
	BYTE *pbIn = (BYTE*)AlignedMalloc(iInSize, 32);
	BYTE *pbOut = (BYTE*)AlignedMalloc(iOutSize * 3 + 1920, 64);
	BYTE *pbOut1 = pbOut;
	BYTE *pbOut2 = pbOut1 + iOutSize;
	BYTE *pbOut3 = pbOut2 + iOutSize;

	BYTE *pbSrcY = pbIn;
	BYTE *pbDstY, *pbDstU, *pbDstV;
	
	printf("pbOut1 %p pbOut2 %p pbOut3 %p\n", pbOut1, pbOut2, pbOut3);
#ifdef WIN32
	LARGE_INTEGER llFreq;
	QueryPerformanceFrequency(&llFreq);
#endif

	//generate src data
	srand((int)time(0));
	for(int j=0; j < iHeight; j++)
	{
		for(int i=0; i<iWidth * 2; i++)
		{
			int val = rand() % 255;
			pbSrcY[j * iWidth * 2 + i] = val;
		}
	}
	//c function
	pbDstY = pbOut1;
	pbDstU = pbDstY + iWidth * iHeight * 2;
	pbDstV = pbDstU + iWidth * iHeight;
	UYVYToYUV422P10DI_C(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);

	//sse2 function
	pbDstY = pbOut2;
	pbDstU = pbDstY + iWidth * iHeight * 2;
	pbDstV = pbDstU + iWidth * iHeight;
	UYVYToYUV422P10DI_xme(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	if(memcmp(pbOut1, pbOut2, iOutSize))
	{
		printf("sse2 result doesn't match\n");
		return 0;
	}

#ifdef _AVX2
	//avx2 function
	pbDstY = pbOut3;
	pbDstU = pbDstY + iWidth * iHeight * 2;
	pbDstV = pbDstU + iWidth * iHeight;
	UYVYToYUV422P10DI_avx2(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	if(memcmp(pbOut1, pbOut3, iOutSize))
	{
		printf("avx2 result doesn't match\n");
		return 0;
	}
#endif

#ifdef _AVX512
	//avx2 function
	pbDstY = pbOut3;
	pbDstU = pbDstY + iWidth * iHeight * 2;
	pbDstV = pbDstU + iWidth * iHeight;
	UYVYToYUV422P10DI_avx512(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	if(memcmp(pbOut1, pbOut3, iOutSize))
	{
		printf("avx512 result doesn't match\n");
		FILE *pFile = fopen("../avx512.yuv", "wb");
		if(pFile)
		{
			fwrite(pbOut3, iOutSize, 1, pFile);
			fclose(pFile);
		}
		pFile = fopen("../c.yuv", "wb");
		if(pFile)
		{
			fwrite(pbOut1, iOutSize, 1, pFile);
			fclose(pFile);
		}
		return 0;
	}
#endif

	INT64 t_cpu, t_sse2, t_avx2, t_avx512;
	INT64 t_start, t_end;
	t_start = GetPerfCount();
	for(int i=0; i < iFrames; i++)
	{
		UYVYToYUV422P10DI_C(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	}
	t_end = GetPerfCount();
	t_cpu = t_end - t_start;

	t_start = GetPerfCount();
	for(int i=0; i < iFrames; i++)
	{
		UYVYToYUV422P10DI_xme(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	}
	t_end = GetPerfCount();
	t_sse2 = t_end - t_start;

	t_start = GetPerfCount();
#ifdef _AVX2
	for(int i=0; i < iFrames; i++)
	{
		UYVYToYUV422P10DI_avx2(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	}
#endif
	t_end = GetPerfCount();
	t_avx2 = t_end - t_start;

	t_start = GetPerfCount();
#ifdef _AVX512
	for(int i=0; i < iFrames; i++)
	{
		UYVYToYUV422P10DI_avx512(pbDstY, pbDstU, pbDstV, iWidth * 2, iWidth, pbSrcY, NULL, NULL, iWidth * 2, 0, iWidth, iHeight);
	}
#endif
	t_end = GetPerfCount();
	t_avx512 = t_end - t_start;

#ifdef WIN32
	printf("time cpu %lld sse2 %lld avx2 %lld\n", t_cpu*1000/iFrames/llFreq.QuadPart, t_sse2*1000/iFrames/llFreq.QuadPart, t_avx2*1000/iFrames/llFreq.QuadPart);
#else
	printf("time cpu %lld sse2 %lld avx2 %lld avx512 %lld\n", t_cpu/iFrames, t_sse2/iFrames, t_avx2/iFrames, t_avx512/iFrames);
#endif


	if(pbIn)
		AlignedFree(pbIn);
	if(pbOut)
		AlignedFree(pbOut);

	return 0;
}

